using Revise
Revise.track(@__FILE__)

using LinearAlgebra: Diagonal, dot

using Colors: HSV
using DualTVDD:
    DualTVL1ROFOpProblem, ChambolleAlgorithm, ProjGradAlgorithm,
    init, step!, fetch, recover_u, gradient
using FileIO
using ImageIO
using Plots
using StaticKernels: Kernel, ExtensionReplicate, extend


function OptFlowProblem(f0::AbstractArray{T,d}, f1::AbstractArray{T,d}, λ, β = 1e-5) where {T,d}
    ∇ = Kernel{ntuple(_->0:1, d)}(gradient)
    ∇f0 = map(∇, extend(f0, ExtensionReplicate()))

    Bval = inv.(∇f0 .* adjoint.(∇f0) .+ β .* Ref(I))
    B = Diagonal(vec(Bval))

    g = .-(f1 .- f0) .* ∇f0

    return DualTVL1ROFOpProblem(g, B, λ)
end

function run_optflow(f0, f1, λ=0.01, β=0.001)
    prob = OptFlowProblem(f0, f1, λ, β)
    #alg = ChambolleAlgorithm(prob, τ=1/10000)
    #alg = ProjGradAlgorithm(prob, τ=1/10000)
    #alg = ProjGradAlgorithm(prob)
    alg = DualTVDD.DualTVDDAlgorithm(prob, M=(2,2), overlap=(4,4))

    ctx = init(alg)
    for i in 1:1000
        step!(ctx)
    end
    plot_optflow(ctx)

    return ctx
end

color(v) = HSV(180 / pi * atan(v[1], v[2]), norm(v), 1)

function plot_optflow(ctx)
    u = recover_u(fetch(ctx), ctx.algorithm.problem)

    flowimg = color.(u)

    plt = plot(flowimg)
    display(plt)
end


f0 = Float64.(load("data/other-data-gray/RubberWhale/frame10.png"))
f1 = Float64.(load("data/other-data-gray/RubberWhale/frame11.png"))
#run_optflow(f0, f1)
